@@ -69,6 +69,11 @@ class AutoIncrementField(AutoField):
         return NodeList((node_list, SQL('AUTOINCREMENT')))
 
 
+class TDecimalField(DecimalField):
+    field_type = 'TEXT'
+    def get_modifiers(self): pass
+
+
 class JSONPath(ColumnBase):
     def __init__(self, field, path=None):
         super(JSONPath, self).__init__()
@@ -88,11 +93,11 @@ class JSONPath(ColumnBase):
 
     def set(self, value, as_json=None):
         if as_json or isinstance(value, (list, dict)):
-            value = fn.json(json.dumps(value))
+            value = fn.json(self._field._json_dumps(value))
         return fn.json_set(self._field, self.path, value)
 
     def update(self, value):
-        return self.set(fn.json_patch(self, json.dumps(value)))
+        return self.set(fn.json_patch(self, self._field._json_dumps(value)))
 
     def remove(self):
         return fn.json_remove(self._field, self.path)
@@ -117,16 +122,37 @@ class JSONPath(ColumnBase):
 class JSONField(TextField):
     field_type = 'JSON'
 
+    def __init__(self, json_dumps=None, json_loads=None, **kwargs):
+        self._json_dumps = json_dumps or json.dumps
+        self._json_loads = json_loads or json.loads
+        super(JSONField, self).__init__(**kwargs)
+
     def python_value(self, value):
         if value is not None:
             try:
-                return json.loads(value)
+                return self._json_loads(value)
             except (TypeError, ValueError):
                 return value
 
     def db_value(self, value):
         if value is not None:
-            return json.dumps(value)
+            if not isinstance(value, Node):
+                value = fn.json(self._json_dumps(value))
+            return value
+
+    def _e(op):
+        def inner(self, rhs):
+            if isinstance(rhs, (list, dict)):
+                rhs = Value(rhs, converter=self.db_value, unpack=False)
+            return Expression(self, op, rhs)
+        return inner
+    __eq__ = _e(OP.EQ)
+    __ne__ = _e(OP.NE)
+    __gt__ = _e(OP.GT)
+    __ge__ = _e(OP.GTE)
+    __lt__ = _e(OP.LT)
+    __le__ = _e(OP.LTE)
+    __hash__ = Field.__hash__
 
     def __getitem__(self, item):
         return JSONPath(self)[item]
@@ -1169,6 +1195,33 @@ def bm25(raw_match_info, *args):
     L_O = A_O + col_count
     X_O = L_O + col_count
 
+    # Worked example of pcnalx for two columns and two phrases, 100 docs total.
+    # {
+    #   p  = 2
+    #   c  = 2
+    #   n  = 100
+    #   a0 = 4   -- avg number of tokens for col0, e.g. title
+    #   a1 = 40  -- avg number of tokens for col1, e.g. body
+    #   l0 = 5   -- curr doc has 5 tokens in col0
+    #   l1 = 30  -- curr doc has 30 tokens in col1
+    #
+    #   x000     -- hits this row for phrase0, col0
+    #   x001     -- hits all rows for phrase0, col0
+    #   x002     -- rows with phrase0 in col0 at least once
+    #
+    #   x010     -- hits this row for phrase0, col1
+    #   x011     -- hits all rows for phrase0, col1
+    #   x012     -- rows with phrase0 in col1 at least once
+    #
+    #   x100     -- hits this row for phrase1, col0
+    #   x101     -- hits all rows for phrase1, col0
+    #   x102     -- rows with phrase1 in col0 at least once
+    #
+    #   x110     -- hits this row for phrase1, col1
+    #   x111     -- hits all rows for phrase1, col1
+    #   x112     -- rows with phrase1 in col1 at least once
+    # }
+
     weights = get_weights(col_count, args)
 
     for i in range(term_count):
@@ -1192,8 +1245,8 @@ def bm25(raw_match_info, *args):
             avg_length = float(match_info[A_O + j]) or 1.  # avgdl
             ratio = doc_length / avg_length
 
-            num = term_frequency * (K + 1)
-            b_part = 1 - B + (B * ratio)
+            num = term_frequency * (K + 1.0)
+            b_part = 1.0 - B + (B * ratio)
             denom = term_frequency + (K * b_part)
 
             pc_score = idf * (num / denom)
